from django_login_email.email import EmailLoginInfo

sample_system_name = "My System"


class MyLoginInfo(EmailLoginInfo):
  system_name = sample_system_name


def test_custom_system_name():
  e = MyLoginInfo()
  assert (
    e.subject == f"Welcome to {sample_system_name}! Please click the link below to login."
  )
  assert e.from_email == "noreply@example.com"



from django.core.management import call_command
from django.core.management.base import BaseCommand


class Command(BaseCommand):
  def handle(self, *args, **kwargs):
    call_command("makemigrations")
    call_command("migrate")



from django import forms
from django.contrib.auth import get_user_model

from . import register


class LoginForm(forms.Form):
  """Just check the login request."""

  email = forms.EmailField(label="email")


class RegisterForm(forms.Form):
  """Just check the register request."""

  email = forms.EmailField(label="email")

  def is_valid(self) -> bool:
    res = super().is_valid()
    if register.is_registered(self.cleaned_data["email"]):
      self.add_error("email", "User already registered.")
      return False
    return res



import datetime

from .. import email, models, token
from . import utils


class MailRecordModelMixin(email.EmailFunc):
  """Here is an example for MailRecord, using django model. You could implement yourself."""

  def reset_mail(self, mail: str):
    """reset mail token expired time."""
    # models.EmailLogin.objects.filter(email=mail).delete()
    e = models.EmailRecord.objects.get(email=mail)
    e.expired_time = e.expired_time - datetime.timedelta(minutes=self.tl.minutes)
    e.validated = False
    e.save()

  def get_mail_record(self, mail: str) -> email.MailRecord:
    """get mail record to validate the sault, and validated status."""
    # for easy to change. use a function.
    try:
      e = models.EmailRecord.objects.get(email=mail)
      return email.MailRecord(
        email=e.email, expired_time=e.expired_time, validated=e.validated, sault=e.sault
      )
    except models.EmailRecord.DoesNotExist:
      return email.MailRecord(email=mail, expired_time=None, validated=False, sault="")

  def transform_timestamp(self, ts: int) -> datetime.datetime:
    return utils.transform_timestamp(ts)

  def save_token(self, token: token.TokenDict):
    """When generate new token, should call this method."""
    try:
      mail = models.EmailRecord.objects.get(email=token["email"])
      mail.sault = token["salt"]
      mail.expired_time = self.transform_timestamp(token["expired_time"])
      mail.mail_type = token["mail_type"]
      mail.validated = False
      mail.save()

    except models.EmailRecord.DoesNotExist:
      models.EmailRecord.objects.create(
        email=token["email"],
        sault=token["salt"],
        expired_time=self.transform_timestamp(token["expired_time"]),
        mail_type=token["mail_type"],
        validated=False,
      )

  def disable_token(self, token: token.TokenDict):
    models.EmailRecord.objects.filter(sault=token["salt"]).update(validated=True)



import datetime

import pytest

from django_login_email.email import EmailFunc, MailRecord, TimeLimit
from django_login_email.token import TokenDict

sample_mail = "svtter@163.com"


class MyTimeLit(TimeLimit):
  minutes = 10


class Mixin(EmailFunc):
  tl = MyTimeLit()

  def get_mail_record(self, mail: str) -> MailRecord:
    r1 = MailRecord(
      email=sample_mail,
      expired_time=datetime.datetime.now(tz=datetime.timezone.utc)
      - datetime.timedelta(minutes=10),
      validated=False,
      sault="",
    )
    return r1

  def save_token(self, token: TokenDict):
    pass

  def disable_mail(self, mail: str):
    pass

  def disable_token(self, token: TokenDict):
    pass


@pytest.mark.django_db
def test_mixin():
  e = Mixin()
  re = e.get_mail_record(sample_mail)
  assert re.expired_time
  assert e.check_could_send(email=sample_mail)


class Mixin2(EmailFunc):
  tl = MyTimeLit()

  def get_mail_record(self, mail: str) -> MailRecord:
    r2 = MailRecord(email=sample_mail, expired_time=None, validated=False, sault="")
    return r2

  def save_token(self, token: TokenDict):
    pass

  def disable_token(self, token: TokenDict):
    pass


def test_none_of_mixin():
  e = Mixin2()
  re = e.get_mail_record(sample_mail)
  assert re.expired_time is None
  assert e.check_could_send(email=sample_mail)



from django.shortcuts import render
from django.urls import reverse

from django_login_email import email as e
from django_login_email import views as v

# Create your views here.

loginInfo, registerInfo = e.get_info_class("meterhub")


class LoginView(v.EmailLoginView):
  login_info_class = loginInfo
  register_info_class = registerInfo


class VerifyView(v.EmailVerifyView):
  def get_success_url(self):
    return reverse("home")


class LogoutView(v.EmailLogoutView):
  pass



# Create your views here.
import logging
from typing import Any

from django.http import HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.views.generic.edit import FormView

from django_login_email import email, forms

from . import limit
from .mixin import MailRecordModelMixin

logger = logging.getLogger(__name__)


class EmailLoginView(FormView, MailRecordModelMixin):
  """process login by email"""

  template_name = "login_email/login.html"
  error_template: str = "login_email/error.html"
  success_template: str = "login_email/success.html"

  form_class = forms.LoginForm

  login_info_class = email.EmailLoginInfo
  register_info_class = email.EmailRegisterInfo

  tl = limit.LoginTimeLimit()

  def get(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
    if self.request.user.is_authenticated:
      return redirect("home")
    return super().get(request, *args, **kwargs)

  def form_valid(self, form):
    """check the email"""

    # Not allow to login if the user is already authenticated.
    if self.request.user.is_authenticated:
      return redirect("home")

    try:
      # send login mail. If user not exist, send register mail.
      self.send_login_mail(form.cleaned_data["email"])
    except Exception as e:
      logger.error(e)
      return render(self.request, self.error_template, {"error": e})
    return render(self.request, self.success_template, {"form": form})



# Create your views here.
import logging
from typing import Any

from django.http import Http404, HttpRequest, HttpResponse
from django.shortcuts import redirect, render
from django.urls import reverse
from django.views.generic import TemplateView

from django_login_email import email, errors

from . import limit
from .login import EmailLoginView  # noqa
from .mixin import MailRecordModelMixin

logger = logging.getLogger(__name__)


class EmailVerifyView(TemplateView, email.EmailVerifyMixin, MailRecordModelMixin):
  """verify token in url"""

  tl = limit.LoginTimeLimit()
  error_template: str = "login_email/error.html"

  def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
    token = request.GET.get("token", None)
    if token is None:
      raise Http404("Invalid Request")
    try:
      self.verify_login_mail(request=request, token_v=token)
    except errors.ValidatedError as e:
      return render(self.request, self.error_template, {"error": e})
    except Exception as e:
      logger.error(e)
      raise Http404("Invalid Request")
    return redirect(self.get_success_url())

  def get_success_url(self):
    return reverse("login_email:home")


class EmailLogoutView(TemplateView, email.EmailLogoutMixin):
  login_url: str = "login_email:login"

  def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
    self.logout(request=request)
    return redirect(self.login_url)


class HomeView(TemplateView):
  template_name = "login_email/home.html"

  def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
    if not self.request.user.is_anonymous and self.request.user.is_authenticated:
      return super().get(request, *args, **kwargs)
    else:
      return redirect("login_email:login")



"""
Django settings for example project.

Generated by 'django-admin startproject' using Django 5.0.4.

For more information on this file, see
https://docs.djangoproject.com/en/5.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/5.0/ref/settings/
"""

import os
from pathlib import Path

from dotenv import find_dotenv, load_dotenv

load_dotenv(find_dotenv())

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/5.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-m+sp=2@325$cjgicu(=-@5y2h5!fq)syl6)4@ucavnt83693(w"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
  "django.contrib.admin",
  "django.contrib.auth",
  "django.contrib.contenttypes",
  "django.contrib.sessions",
  "django.contrib.messages",
  "django.contrib.staticfiles",
  "django_login_email",
  "testapp",
]

MIDDLEWARE = [
  "django.middleware.security.SecurityMiddleware",
  "django.contrib.sessions.middleware.SessionMiddleware",
  "django.middleware.common.CommonMiddleware",
  "django.middleware.csrf.CsrfViewMiddleware",
  "django.contrib.auth.middleware.AuthenticationMiddleware",
  "django.contrib.messages.middleware.MessageMiddleware",
  "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

ROOT_URLCONF = "settings.urls"

TEMPLATES = [
  {
    "BACKEND": "django.template.backends.django.DjangoTemplates",
    "DIRS": [],
    "APP_DIRS": True,
    "OPTIONS": {
      "context_processors": [
        "django.template.context_processors.debug",
        "django.template.context_processors.request",
        "django.contrib.auth.context_processors.auth",
        "django.contrib.messages.context_processors.messages",
      ],
    },
  },
]

WSGI_APPLICATION = "settings.wsgi.application"


# Database
# https://docs.djangoproject.com/en/5.0/ref/settings/#databases


database_pg = {
  "ENGINE": "django.db.backends.postgresql",
  "NAME": "example2",
  "USER": "postgres",
  "PASSWORD": "example",
  "HOST": "localhost",
  "PORT": "5432",
}

database_sqlite = {
  "ENGINE": "django.db.backends.sqlite3",
  "NAME": BASE_DIR / "db.sqlite3",
}

DATABASES = {
  "default": database_sqlite,
}


# Password validation
# https://docs.djangoproject.com/en/5.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
  {
    "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
  },
  {
    "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
  },
  {
    "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
  },
  {
    "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
  },
]


# Internationalization
# https://docs.djangoproject.com/en/5.0/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.0/howto/static-files/

STATIC_URL = "static/"

# Default primary key field type
# https://docs.djangoproject.com/en/5.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


# You could test this by using https://mailtrap.io/
# EMAIL_HOST = "sandbox.smtp.mailtrap.io"
# EMAIL_HOST_USER = "f9c15aa2704df2"
# EMAIL_HOST_PASSWORD = os.getenv("EMAIL_HOST_PASSWORD")
# EMAIL_PORT = "2525"

EMAIL_HOST = "localhost"
EMAIL_HOST_USER = "user"
EMAIL_HOST_PASSWORD = os.getenv("EMAIL_HOST_PASSWORD", "password")
EMAIL_PORT = "1025"

# Logging configuration
LOGGING = {
  "version": 1,
  "disable_existing_loggers": False,
  "formatters": {
    "verbose": {
      "format": "{levelname} {asctime} {module} {process:d} {thread:d} {message}",
      "style": "{",
    },
    "simple": {
      "format": "{levelname} {message}",
      "style": "{",
    },
  },
  "handlers": {
    "console": {
      "class": "logging.StreamHandler",
      "formatter": "verbose",
    },
    "file": {
      "class": "logging.FileHandler",
      "filename": "debug.log",
      "formatter": "verbose",
    },
  },
  "root": {
    "handlers": ["console", "file"],
    "level": "INFO",
  },
  "loggers": {
    "django": {
      "handlers": ["console", "file"],
      "level": "INFO",
      "propagate": True,
    },
  },
}



from django.urls import path

from . import views as v

app_name = "login_email"

urlpatterns = [
  path("login", v.LoginView.as_view(), name="login"),
  path("verify", v.VerifyView.as_view(), name="verify"),
  path("logout", v.LogoutView.as_view(), name="logout"),
]



# Generated by Django 5.0.4 on 2024-04-06 03:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_login_email', '0002_emaillogin_sault_emaillogin_validated_and_more'),
    ]

    operations = [
        migrations.RenameField(
            model_name='emaillogin',
            old_name='last_time',
            new_name='expired_time',
        ),
    ]



def test_case_1():
  # TODO: functional test.
  """
  Sending email, validate token, and login, success
  """

  """
  Sending email, wait for 10 minute, login, failed,
  """

  """
  Sending email, wait for 10 minute, sending email again, validate token, and login, success
  """



# Generated by Django 5.0.4 on 2024-04-06 03:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_login_email', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='emaillogin',
            name='sault',
            field=models.CharField(default='', max_length=100, verbose_name='Sault'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='emaillogin',
            name='validated',
            field=models.BooleanField(default=False, verbose_name='Login Token validated'),
        ),
        migrations.AlterField(
            model_name='emaillogin',
            name='email',
            field=models.EmailField(max_length=254, unique=True, verbose_name='Email'),
        ),
    ]



from django.contrib.auth import get_user_model


def test_email_mixin(db, mx_send):
  mr = mx_send.get_mail_record("svtter@163.com")
  assert mr.expired_time is None


def test_is_send(db, mx_send):
  is_call = False

  def fake_fn(*arg, **kwargs):
    nonlocal is_call
    is_call = True

  mx_send.save_token = fake_fn
  mx_send.send_valid("svtter@163.com", "login")

  assert is_call


def test_send_valid(db, mx_send):
  mx_send.send_valid("svtter@163.com", "login")

  # check mail record
  mr = mx_send.get_mail_record("svtter@163.com")
  assert mr.expired_time is not None
  assert mr.validated is False


def test_user_exist(db, mx_send):
  User = get_user_model()
  User.objects.create_user(username="svtter", email="svtter@163.com", password="123456")

  mx_send.send_valid("svtter@163.com", "login")
  mr = mx_send.get_mail_record("svtter@163.com")
  assert mr.validated is False



import pytest
from django.contrib.auth import get_user_model

from django_login_email import email, token
from django_login_email.views import mixin


class MyVerify(email.EmailVerifyMixin, mixin.MailRecordModelMixin):
  tl = email.TimeLimit(10)


@pytest.fixture
def mx_verify():
  return MyVerify()


def wrap_token_manager(mx_send):
  token_b = None

  class TokenManager(token.TokenManager):
    def encrypt_mail(self, *args, **kwargs):
      nonlocal token_b
      token_b = super().encrypt_mail(*args, **kwargs)
      return token_b

  def get_token_manager():
    return TokenManager(10)

  def get_token_b():
    return token_b

  mx_send.get_token_manager = get_token_manager
  return get_token_b


def test_verify(db, mx_send, mx_verify):
  """test token validate"""
  # no user
  get_token_b = wrap_token_manager(mx_send)
  mx_send.send_valid("svtter@163.com", "login")

  u = mx_verify.verify_token(get_token_b())
  assert u.email == "svtter@163.com"


def test_has_user(db, mx_send, mx_verify):
  # exist user
  User = get_user_model()
  User.objects.create(email="svtter@163.com", password="123456")

  # verify token
  get_token_b = wrap_token_manager(mx_send)
  mx_send.send_valid("svtter@163.com", "login")
  mx_verify.verify_token(get_token_b())

  mr = mx_send.get_mail_record("svtter@163.com")
  assert mr.validated

  assert User.objects.filter(email="svtter@163.com").exists()


def test_login_twice(db, mx_send, mx_verify):
  User = get_user_model()
  User.objects.create_user(username="svtter", email="svtter@163.com", password="123456")

  # first login
  get_token_b = wrap_token_manager(mx_send)
  mx_send.send_valid("svtter@163.com", "login")
  mr = mx_send.get_mail_record("svtter@163.com")
  assert not mr.validated

  mx_verify.verify_token(get_token_b())
  mr = mx_send.get_mail_record("svtter@163.com")
  assert mr.validated

  # second login
  get_token_b = wrap_token_manager(mx_send)
  mx_send.send_valid("svtter@163.com", "login")
  mr = mx_send.get_mail_record("svtter@163.com")
  assert not mr.validated

  mx_verify.verify_token(get_token_b())
  mr = mx_send.get_mail_record("svtter@163.com")
  assert mr.validated



import pytest

from django_login_email import email
from django_login_email.views import mixin

loginInfo, registerInfo = email.get_info_class("system")


class MixinTest(mixin.MailRecordModelMixin):
  login_info_class = loginInfo
  register_info_class = registerInfo
  tl = email.TimeLimit()


@pytest.fixture
def mx_send():
  return MixinTest()



# Generated by Django 5.1.4 on 2025-01-14 11:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_login_email', '0003_rename_last_time_emaillogin_expired_time'),
    ]

    operations = [
        migrations.CreateModel(
            name='EmailRegister',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('expired_time', models.DateTimeField(auto_now_add=True, verbose_name='Last login request time')),
                ('validated', models.BooleanField(default=False, verbose_name='Login Token validated')),
                ('sault', models.CharField(max_length=100, verbose_name='Sault')),
                ('email', models.EmailField(max_length=254, unique=True, verbose_name='Email')),
            ],
        ),
    ]



from django.contrib.auth import get_user_model


def is_registered(email: str) -> bool:
  """
  check if the user is already registered.
  If user is not active, return False.
  """
  User = get_user_model()
  u = User.objects.filter(email=email)
  if u.exists():
    user = u.first()
    if user.is_active:
      return True
  return False



from django.core.management.base import BaseCommand

from django_login_email import models


class Command(BaseCommand):
  def handle(self, *args, **kwargs):
    print(models.EmailRecord.objects.all())



from django.contrib import admin

from django_login_email import models

# Register your models here.


@admin.register(models.EmailRecord)
class EmailRecordAdmin(admin.ModelAdmin):
  list_display = ("email", "expired_time", "validated", "mail_type")



# Handle email token
import base64
import datetime
import json
import logging
import os
import typing as t
import urllib.parse

from Crypto.Cipher import AES
from django.conf import settings

Token = str
EmailAndSalt = str
Email = str

TokenDict = t.TypedDict(
  "TokenDict", {"email": Email, "expired_time": int, "salt": str, "mail_type": str}
)

logger = logging.getLogger(__name__)


class TokenGenerator(object):
  """generate token by minutes"""

  def __init__(self, minutes: int) -> None:
    self.minutes = minutes

  def get_expired_time(self) -> int:
    return int(
      (datetime.datetime.now() + datetime.timedelta(minutes=self.minutes)).timestamp()
    )

  def gen_salt(self, email) -> str:
    return email + str(base64.b64encode(os.urandom(16)))

  def gen(self, email, mail_type: str, save_token: t.Callable[[TokenDict], None]) -> str:
    """call save_token to save token in database or somewhere"""
    token: TokenDict = {
      "email": email,
      "expired_time": self.get_expired_time(),
      "salt": self.gen_salt(email),
      "mail_type": mail_type,
    }
    save_token(token)
    token_str = json.dumps(token)
    return token_str


class TokenManager(object):
  """manage the email generate token"""

  key: bytes
  generator: TokenGenerator

  def __init__(self, minutes: int) -> None:
    self.key = settings.SECRET_KEY[:32].encode("utf-8")
    self.generator: TokenGenerator = TokenGenerator(minutes)

  def transform_token(self, token_uncrypt: Token) -> TokenDict:
    token_dict: TokenDict = json.loads(token_uncrypt)
    return token_dict

  def check_token(
    self, token_dict: TokenDict, get_salt: t.Callable[[], str]
  ) -> t.Optional[TokenDict]:
    """check salt and expire-time"""
    logger.info(f"token_dict: {token_dict}")
    if not token_dict["salt"] == get_salt():
      logger.info(f"salt is error, {token_dict['salt']}, {get_salt()}")
      return None
    if token_dict["expired_time"] > int(datetime.datetime.now().timestamp()):
      logger.info("expired_time is ok")
      return token_dict
    logger.info("expired_time is error")
    return None

  def get_mail(self, token_uncrypt: TokenDict) -> Email:
    """Read the email from token"""
    return token_uncrypt["email"]

  def _encrypt(self, plaintext, key):
    cipher = AES.new(key, AES.MODE_EAX)
    nonce = cipher.nonce
    ciphertext, tag = cipher.encrypt_and_digest(plaintext)
    return nonce + ciphertext + tag

  def _decrypt(self, ciphertext, key):
    nonce = ciphertext[:16]
    tag = ciphertext[-16:]
    ciphertext = ciphertext[16:-16]
    cipher = AES.new(key, AES.MODE_EAX, nonce)
    plaintext = cipher.decrypt_and_verify(ciphertext, tag)
    return plaintext

  def encrypt_mail(
    self, email: Email, mail_type: str, save_token: t.Callable[[TokenDict], None]
  ) -> str:
    """email -> token -> base64 -> utf-8 -> urlquote"""
    # add salt.
    content = self.generator.gen(email, mail_type, save_token)
    content = content.encode("utf-8")
    return urllib.parse.quote(
      base64.b64encode(self._encrypt(content, self.key)).decode("utf-8")
    )

  def decrypt_token(self, token: Token) -> str:
    t = urllib.parse.unquote(token).encode("utf-8")
    t = base64.b64decode(t)
    return self._decrypt(t, self.key).decode("utf-8")



class LoginMailError(Exception):
  pass


class ValidatedError(LoginMailError):
  """When validated, raise this error"""

  pass


class TokenError(LoginMailError):
  """When token error, raise this error"""

  pass



from django.db import models

# Create your models here.


class EmailRecord(models.Model):
  """Record the token for login/register."""

  expired_time = models.DateTimeField(
    auto_now_add=True, verbose_name="Last register request time"
  )
  validated = models.BooleanField(default=False, verbose_name="Register Token validated")
  mail_type = models.CharField(max_length=100, verbose_name="Mail type")
  sault = models.CharField(max_length=100, verbose_name="Sault")
  email = models.EmailField(verbose_name="Email", unique=True, null=False)

  @classmethod
  def set_email_expired_time(cls, email, datetime) -> "EmailRecord":
    obj, _ = cls.objects.get_or_create(email=email)
    obj.set_expired_time(datetime)
    return obj

  def set_expired_time(self, datetime):
    self.expired_time = datetime
    self.validated = False
    self.save()

  def __str__(self) -> str:
    return f"Email: {self.email}, mail_type: {self.mail_type}"



import abc
import datetime
import string
import typing as t
from dataclasses import dataclass
from datetime import timezone

from django.contrib.auth import get_user_model, login, logout
from django.core.mail import EmailMessage

from . import errors, token


class EmailInfo(object):
  """Email info."""

  subject: str
  message: str
  from_email: str
  welcome_text: str
  system_name: str

  url: str = "http://127.0.0.1:8000/account/verify?token="
  login_message: string.Template = string.Template('Click <a href="$url$token">Link</a>')

  def set_token(self, value):
    self.message = self.welcome_text + self.login_message.substitute(
      url=self.url, token=value
    )


class EmailLoginInfo(EmailInfo):
  """Email info for login."""

  def __init__(self):
    self.subject = f"Welcome to {self.system_name}! Please click the link below to login."
    self.from_email = "noreply@example.com"
    self.welcome_text: str = (
      f"Welcome to {self.system_name}! Please click the link below to login.<br>"
    )


class EmailRegisterInfo(EmailInfo):
  """Email info for register."""

  def __init__(self):
    self.subject = (
      f"Welcome to {self.system_name}! Please click the link below to register."
    )
    self.from_email = "noreply@example.com"
    self.welcome_text: str = (
      f"Welcome to {self.system_name}! Please click the link below to register.<br>"
    )


def get_info_class(sys_name: str) -> t.Tuple[EmailLoginInfo, EmailRegisterInfo]:
  class MyLoginInfo(EmailLoginInfo):
    system_name = sys_name

  class MyRegisterInfo(EmailRegisterInfo):
    system_name = sys_name

  return MyLoginInfo, MyRegisterInfo


@dataclass
class TimeLimit(object):
  minutes: int = 10


@dataclass
class MailRecord(object):
  expired_time: t.Optional[datetime.datetime]
  email: str
  validated: bool
  sault: str


class MailRecordAPI(abc.ABC):
  """Mixin to get mail record."""

  @abc.abstractmethod
  def get_mail_record(self, mail: str) -> MailRecord:
    """Get the mail record from the database."""
    raise NotImplementedError("You must implement get_mail_record")

  @abc.abstractmethod
  def save_token(self, token: token.TokenDict):
    """to save token in the database or somewhere."""
    raise NotImplementedError("You must implement save_token")

  @abc.abstractmethod
  def disable_token(self, token: token.TokenDict):
    """Every token should only login once"""
    raise NotImplementedError("")


class EmailFunc(MailRecordAPI):
  """Mixin to send email."""

  login_info_class: t.Type[EmailLoginInfo]
  register_info_class: t.Type[EmailRegisterInfo]

  tl: TimeLimit

  def check_user(self, email) -> bool:
    """check if the user exists."""
    User = get_user_model()
    u = User.objects.filter(email=email)
    return u.exists()

  def check_could_send(self, email) -> bool:
    """check if the email could send."""
    re = self.get_mail_record(email)
    # TODO: if other user send email, the current user could not sign in.
    if (re.expired_time is None) or (
      re.expired_time <= datetime.datetime.now(tz=timezone.utc)
    ):
      return True
    return False

  def get_token_manager(self) -> token.TokenManager:
    return token.TokenManager(self.tl.minutes)

  def send_valid(self, email: str, mail_type: str):
    """send login/register mail."""
    if mail_type == "login":
      e = self.login_info_class()
    elif mail_type == "register":
      e = self.register_info_class()
    else:
      raise ValueError(f"Invalid mail type: {mail_type}")

    m = self.get_token_manager()
    encrypt_token = m.encrypt_mail(email, mail_type, self.save_token)
    e.set_token(encrypt_token)

    msg = EmailMessage(e.subject, e.message, e.from_email, [email])
    msg.content_subtype = "html"
    msg.send()

  def send_login_mail(self, email: str):
    """
    send login mail.
    """
    # if user not exist, send register mail.
    if not self.check_user(email):
      mail_type = "register"
    else:
      mail_type = "login"

    # if the email could not send, raise exception.
    if not self.check_could_send(email=email):
      raise Exception(f"Cannot send. Wait {self.tl.minutes} minutes.")

    self.send_valid(email, mail_type)


class EmailVerifyMixin(MailRecordAPI):
  """verify the token in url"""

  tl: TimeLimit

  def verify_token(self, token_v: str):
    m = token.TokenManager(self.tl.minutes)
    token_str = m.decrypt_token(token=token_v)
    token_d = m.transform_token(token_str)

    mr = self.get_mail_record(m.get_mail(token_d))
    if mr.validated:
      raise errors.ValidatedError("Token already validated.")

    token_d = m.check_token(token_d, lambda: mr.sault)
    if token_d is None:
      raise errors.TokenError("Invalid token.")

    User = get_user_model()
    u = User.objects.filter(email=m.get_mail(token_d)).first()
    if not u:
      # if user not exist, create a new user.
      # support register by email.
      u = User.objects.create(username=m.get_mail(token_d), email=m.get_mail(token_d))

    if not u.is_active:
      raise Exception("Inactive user, disallow login.")

    self.disable_token(token=token_d)
    return u

  def verify_login_mail(self, request, token_v: str):
    """
    verify the login mail.
    if user not exist, create a new user.
    """
    u = self.verify_token(token_v=token_v)
    login(request, u)


class EmailLogoutMixin(object):
  def logout(self, request):
    logout(request)



from django_login_email import email


class MyLoginInfo(email.EmailLoginInfo):
  system_name = "meterhub"


def test_email_info():
  e = MyLoginInfo()
  e.set_token("1234567890")
  assert (
    e.message
    == """Welcome to meterhub! Please click the link below to login.<br>Click <a href="http://127.0.0.1:8000/account/verify?token=1234567890">Link</a>"""
  )



from typing import Callable

from django_login_email.forms import register


def raise_if_registered(func: Callable[[str], str]):
  """Raise an exception if the user is already registered."""

  # check params of func
  assert "email" in func.__code__.co_varnames, "Function has no 'email' parameter."

  def inner(email: str):
    if register.is_registered(email):
      raise Exception("User already registered.")
    return func(email)

  return inner



from django.contrib.auth.models import User
from django.core.management.base import BaseCommand


class Command(BaseCommand):
  def handle(self, *args, **kwargs):
    User.objects.create_superuser(
      username="test@test.com",
      email="test@test.com",
      password="test",
    )
    self.stdout.write("Create superuser test@test.com, with password test.")



import datetime

import pytest

from django_login_email import models


@pytest.mark.django_db
def test_set_expired_time():
  now_t = datetime.datetime.now()
  e = models.EmailRecord.set_email_expired_time(email="svtter@163.com", datetime=now_t)
  assert models.EmailRecord.objects.get(email="svtter@163.com")

  now_t2 = datetime.datetime.now() + datetime.timedelta(minutes=10)
  e.set_expired_time(now_t2)
  assert e.expired_time == now_t2
